/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <regex.h>

#include "lkpatch.h"

int streq(char*s1,char*s2) {
  return(strcmp(s1,s2)==0);
}

char*cfname(void) { /* rtn name of configurationfile */
 static char*cf;
 struct passwd *p;
  if(cf) return(cf);
  if(!(p=getpwuid(getuid()))) error(1,errno,"cfname: can't get passwd-entry");
  if(!(cf=malloc(strlen(p->pw_dir)+1+strlen(CF)+1))) {
    error(1,errno,"cfname: out of mem");
  }
  strcpy(cf,p->pw_dir); strcat(cf,"/"); strcat(cf,CF);
  return(cf);
}

int cfthere(void) {
 char*cf=cfname();
 struct stat sbuf;
  if(stat(cf,&sbuf)) return(0);
  if(sbuf.st_size==0) return(0);
  if(access(cf,R_OK|W_OK)) error(1,0,"no permissions to conf file %s",cf);
  return(1);
}

void cfwrite(UBUF*ub) { /* write config file */
 UBUFE*ue;
 FILE*fp;
 int i,n;
 char*cf=cfname();
  if((fp=fopen(cf,"w"))==NULL) {
    error(1,errno,"cfwrite: can't open %s for writing",cf);
  }
  n=ublen(ub);
  for(i=0;i<n;i++) {
    ue=ubel(ub,i);
    dp(DALL,"cfwrite: \"%s %s %ld\"",ue->url,ue->vhost,ue->q);
    fprintf(fp,"%s %s %ld\n",ue->url,ue->vhost,ue->q);
  }
  if(fclose(fp)) error(1,errno,"cfwrite: can't close %s",cf);
}

UBUF*cfread(void) { /* read the configuration file */
 FILE*fp;
 UBUF*ub; UBUFE e;
 int n;
 char*cf=cfname();
  dp(DTRC,"readcf()");
  if(!(fp=fopen(cf,"r"))) error(1,errno,"can't open %s",cf);
  ub=ubnew();
  for(;;) {
    memset(&e,0,sizeof(e));
    if((n=fscanf(fp,"%as %as %ld",&e.url,&e.vhost,&e.q))==EOF&&feof(fp)) break;
    dp(DALL,"cfread: url: %s, vhost: %s, quality: %ld",e.url,e.vhost,e.q);
    if(n!=3) error(1,0,"bad syntax in configuration file");
    ubaddent(ub,&e);
  }
  if(fclose(fp)) error(1,errno,"can't close %s",cf);
  return(ub);
}

static int cfurl(UBUF*ub,char*url) { /* rtn url-index in ub */
 int i;
  for(i=0;i<ublen(ub);i++) {
    if(!strcmp(ubel(ub,i)->url,url)) break;
  }
  if(i>=ublen(ub)) error(1,0,"cfurl: no such url: %s",url);
  return(i);
}

void cfupd(char*url,long q) { /* update cf by adding q to url */
 UBUF*ub=cfread();
 int i=cfurl(ub,url); 
 UBUFE*ue=ubel(ub,i);
  dp(DTRC,"cfupd: updating %s with %ld from %ld to %ld",
    url,q,ue->q,ue->q==-1?q:(long)(0.8*ue->q+0.2*q));
  ue->q=ue->q==-1?q:(long)(0.8*ue->q+0.2*q);
  if(ue->q==0&&streq(ue->vhost,"www.kernel.org")) { ue->q=1; }
  if(ue->q==0) {
    dp(DPRG,"deleting url %s from configuration file",url);
    ubdelel(ub,i);
  }
  cfwrite(ub);
  ubdel(ub);
}

void cfpen(char*url,double pen) { /* give url a penalty */
 UBUF*ub=cfread();
 int i=cfurl(ub,url);
 UBUFE*ue=ubel(ub,i);
  if(ue->q<=0) ue->q=0;
  ue->q=(long)((double)ue->q*(1-pen));
  dp(DTRC,"cfpen: url %s penalized to %ld",url,ue->q);
  if(ue->q==0&&streq(ue->vhost,"www.kernel.org")) { ue->q=1; }
  if(ue->q==0) {
    dp(DPRG,"deleting url %s from configuration file",url);
    ubdelel(ub,i);
  }
  cfwrite(ub);
  ubdel(ub);
}

void cfdel(char*url) { /* delete url from conf file */
 UBUF*ub=cfread();
 int i=cfurl(ub,url);
  if(streq(ubel(ub,i)->vhost,"www.kernel.org")) return;
  ubdelel(ub,i);
  cfwrite(ub);
  ubdel(ub);
}

int cfsel1(UBUF*ub) { /* return random with q==-1 */
 int l=ublen(ub),i,n,r;
  for(n=i=0;i<l;i++) if(ubel(ub,i)->q==-1) n++;
  r=1+(int)((double)n*rand()/(RAND_MAX+1.0));
  for(i=0;i<l;i++) {
    if(ubel(ub,i)->q==-1) r--;
    if(!r) return(i);
  }
  /* never reached */
  return(-1);
}

int cfsel(UBUF*ub) { /* select a mirror */
 UBUFE*pe;
 int i,n=ublen(ub);
 double p,s,max;
  dp(DTRC,"cfsel()");
  if(n<=0) return(-1);
  for(max=i=0;i<n;i++) {
    pe=ubel(ub,i);
    if(pe->q>max) max=pe->q;
  }
  dp(DALL,"cfsel: max=%ld",max);
  for(i=s=0;i<n;i++) {
   double w;
    dp(DALL,"cfsel: loop1: %d, url: %s",i,ubel(ub,i)->url);
    pe=ubel(ub,i);
    if(pe->q==-1) return(cfsel1(ub));
    w=(double)(pe->q)/max;
    s+=w*w;
  }
  dp(DALL,"cfsel: s=%lf",s);
  srand(time(NULL));
  p=((s*(double)rand())/(RAND_MAX+1.0));
  for(i=s=0;i<n;i++) {
   double w;
    pe=ubel(ub,i);
    dp(DALL,"cfsel: loop2: nr:%d, p:%lf, s:%lf, q:%ld",i,p,s,pe->q);
    w=(double)(pe->q)/max;
    s+=w*w;
    if(p<=s) return(i);
  }
  error(1,0,"cfsel: never reached");
  return(0); /* to keep -Wall happy */
}

/*-----------------------build conf file-----------------------------*/

char*mirurl="http://www.kernel.org/mirrors/";
/*char*mirurl="http://127.0.0.1/mirrors.html";*/
BUF*mirbuf;
char*sitere="(http|ftp)://[^ ]*/pub/linux/";

UBUF*grepsites(BUF*mirlist) {
 regex_t preg;
 regmatch_t match[1];
 UBUF*sites=ubnew();
 char*p;
  dp(DTRC,"grepsites()");
  regcomp(&preg,sitere,REG_EXTENDED|REG_NEWLINE);
  for(p=bufptr(mirlist);!regexec(&preg,p,1,match,0);p+=match[0].rm_eo) {
    p[match[0].rm_eo]='\0';
    dp(DALL,"grepsites: add %s",p+match[0].rm_so);
    ubaddurl(sites,p+match[0].rm_so);
    p[match[0].rm_eo]='/';
  }
  return(sites);
}

UBUF*getmirrors(void) {
 UBUF*rv;
  printf("retrieving mirrors from %s\n",mirurl);
  if((mirbuf=newbuf())==NULL) {
    error(1,errno,"can't allocate mirror buffer");
  }
  if(dlget(mirurl,NULL,mirbuf,dlto)) {
    error(1,0,"download of %s failed: %s",mirurl,dlerr());
  }
  addbuf(mirbuf,"\0",1);
  rv=grepsites(mirbuf);
  delbuf(mirbuf);
  return(rv);
}

void cfmake(void) { /* make config file */
 UBUF*sites;
 UBUFE ue;
 int i;
  sites=getmirrors();
  presel(sites);
  memset(&ue,0,sizeof(ue));
  ue.url=strdup("http://www.kernel.org/pub/linux/");
  ue.vhost=strdup("www.kernel.org");
  ubaddent(sites,&ue);
  for(i=ublen(sites);i>=0;i--) { /* set quality to -1 */
    ubel(sites,i)->q=-1;
  }
  cfwrite(sites);
  ubdel(sites);
}

/*----------------------------------------------------------------------*/

void cfurladd(char*url) { /* add new url to cf */
 UBUF*ub=cfread();
 UBUFE ue;
  ue.url=url;
  ue.vhost=urlhost(url);
  ue.q=-1;
  ubaddent(ub,&ue);
  cfwrite(ub);
}
