/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "lkpatch.h"

int ntries=1;                   /* how many mirrors to try */
int dlto=180;                   /* download timeout */
char*srcdir="/usr/src/linux";   /* linux source dir */

/*------------------------------help-------------------------------------*/

char*hmsg[]={
"usage: lkpatch [options]",
"  options:",
"    -a url: add url to config file"
"    -h    : this help",
"    -i    : insist. Try several mirrors to find a new version.",
"            May be repeated for even more insistence.",
"    -q    : query mode: show installed and current versions.",
"    -R    : Rebuild configuration file",
"    -s dir: kernel source dir",
"    -T t  : Time out downloads if nothing is received for time t.",
"            t may contain 'h','m','s', e.g. 3m10s. Defaults to 3m.",
"    -v nr : verbosity level, default is 1",
"    -V    : print version and exit",
0
};

void help(void) {
 char**p;
  for(p=hmsg;*p;p++) printf("%s\n",*p);
}

/*---------------------------------------------------------------------*/

void pargs(int ac,char**av) {
 int c;
  while((c=getopt(ac,av,"a:Vv:hiqRs:T:"))!=-1) {
    switch(c) {
      case 'a': cfurladd(optarg); break;
      case 'v': setdlvl(atoi(optarg)); break;
      case 'h': help(); exit(0);
      case 'i': ntries*=2; break;
      case 'q': query(); exit(0);
      case 'R': cfmake(); exit(0);
      case 's': srcdir=optarg; break;
      case 'T': dlto=stosec(optarg); break;
      case 'V': printf("lkpatch-0.2.2 (%s)\n",curl_version()); exit(0);
      case '?': error(1,0,"bad option %c. Try -h for help.",optopt);
      default:  error(1,0,"error parsing args");
    }
  }
}

int main(int ac,char**av) {
  assert(sizeof(u32)==4);
  srand(time(NULL));
  pargs(ac,av);
  if(!cfthere()) cfmake();
  patch();
  exit(0);
}
