/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <regex.h>
#include <sys/stat.h>
#include <unistd.h>

#include "lkpatch.h"

void verchk(int v[4]) { /* heuristic version nr check */
 int i;
  for(i=0;i<4;i++) {
    if(i<0||i>200)
      error(1,0,"verchk: bad version %d.%d.%d.%d",v[0],v[1],v[2],v[3]);
  }
}

int verlt(int v1[4],int v2[4]) { /* v1 < v2 ? */
 int i;
  verchk(v1); verchk(v2);
  for(i=0;i<4;i++) {
    if(v1[i]<v2[i]) return(1);
  }
  return(0);
}

int verobs(int o[4],int n[4]) { /* v1 obsolete */
 int i;
  verchk(o); verchk(n);
  if(verlt(o,n)) return(0);
  for(i=0;i<2;i++) {
    if(o[i]!=n[i]) error(1,0,"verobs: versions differ in major levels");
  }
  return(o[2]-n[2]>2);
}

void vercp(int d[4],int s[4]) { /* copy s to d */
  verchk(s);
  memcpy(d,s,4*sizeof(d[0]));
}

void verzero(int v[4]) {
 int i;
  for(i=0;i<4;i++) v[i]=0;
}

void getiver(char*kdir,int v[4]) {
 char*f;
 FILE*fp;
 struct stat sbuf;
  f=emalloc(strlen(kdir)+strlen("/Makefile")+1);
  sprintf(f,"%s/Makefile",kdir);
  if(stat(f,&sbuf)) error(1,0,"%s does not seem to be kernel src dir",kdir);
  if(!(fp=fopen(f,"r"))) error(1,errno,"getiver: can't open %s",f);
  if(fscanf(fp,"VERSION = %d\n",&v[0])!=1) {
    error(1,errno,"getiver: can't read version");
  }
  if(fscanf(fp,"PATCHLEVEL = %d\n",&v[1])!=1) {
    error(1,errno,"getiver: can't read patchlevel");
  }
  if(fscanf(fp,"SUBLEVEL = %d\n",&v[2])!=1) {
    error(1,errno,"getiver: can't read sublevel");
  }
  if(fscanf(fp,"EXTRAVERSION = %d\n",&v[3])!=0) {
    error(1,errno,"getiver: can't deal with extraversions");
  }
  if(fclose(fp)) error(1,errno,"can't close %s",f);
}

void getlv1(BUF*b,int v[4]) { /* get latest version from LATEST-IS symlink */
 char*re="(LATEST-IS-[0-9]+\\.[0-9]+\\.[0-9]+)",*p;
 regex_t preg;
 regmatch_t match;
  dp(DTRC,"getlv1()");
  if(regcomp(&preg,re,REG_EXTENDED|REG_NEWLINE)) {
    error(1,0,"getlv1: regcomp failed");
  }
  verzero(v);
  for(p=bufptr(b);!regexec(&preg,p,1,&match,0);p+=match.rm_eo) { /* match */
   int v2[4];
    v2[3]=0;
    if(sscanf(p+match.rm_so,"LATEST-IS-%d.%d.%d",&v2[0],&v2[1],&v2[2])!=3) {
      error(1,0,"getlv1: can't sscanf latest version");
    }
    dp(DALL,"getlv1: found version %d.%d.%d",v2[0],v2[1],v2[2]);
    if(verlt(v,v2)) vercp(v,v2);
  }
  if(!v[0]) {
    error(1,0,"getlv1: mirror does not supply latest version");
  }
  dp(DTRC,"getlv1: latest version is %d.%d.%d",v[0],v[1],v[2]);
}

#define NEL(x) (sizeof(x)/sizeof(x[0]))

void getlv2(BUF*b,int v[4]) { /* get latest version from files */
 char*p,*re[2];
 regex_t preg;
 regmatch_t match[10];
 int i;
  re[0]="linux-([0-9]+\\.[0-9]+\\.[0-9]+)\\.tar\\.(bz2|gz)";
  re[1]="patch-([0-9]+\\.[0-9]+\\.[0-9]+)\\.(bz2|gz)";
  verzero(v);
  for(i=0;i<2;i++) {
    if(regcomp(&preg,re[i],REG_EXTENDED|REG_NEWLINE)) {
      error(1,0,"getlv2: regcomp failed");
    }
    for(p=bufptr(b);!regexec(&preg,p,NEL(match),match,0);p+=match[2].rm_eo) {
     int v2[4];
      verzero(v2);
      dp(DALL,"getlv2: found version %.*s",
        match[1].rm_eo-match[1].rm_so,p+match[1].rm_so);
      if(sscanf(p+match[1].rm_so,"%d.%d.%d",&v2[0],&v2[1],&v2[2])!=3) {
        error(1,0,"getlv2: can't sscanf version");
      }
      dp(DALL,"getlv2: found version %d.%d.%d",v[0],v[1],v[2]);
      if(verlt(v,v2)) vercp(v,v2);
    }
  }
  dp(DTRC,"getlv2: latest version is %d.%d.%d",v[0],v[1],v[2]);
}

void getlv(BUF*ib,int ver[4]) {
 int v[4];
  verzero(ver);
  getlv1(ib,v); if(verlt(ver,v)) vercp(ver,v);
  getlv2(ib,v); if(verlt(ver,v)) vercp(ver,v);
}

static char*name(char*pre,int v[4],char*suf) {
 char*s;
  s=emalloc(strlen(pre)+strlen("-xxxx.yyyy.zzzz.")+strlen(suf)+1);
  sprintf(s,"%s-%d.%d.%d.%s",pre,v[0],v[1],v[2],suf);
  dp(DALL,"name(%s,%d.%d.%d,%s): %s",pre,v[0],v[1],v[2],suf,s);
  return(s);
}

int inindex(BUF*b,char*ndl) {
 char*hay=bufptr(b);
  dp(DALL,"inindex: looking for %s",ndl);
  return(strstr(hay,ndl)!=0);
}

char*lkurl(char*lxurl,int v[4],char*suf) {
 char*s,*slash;
  s=emalloc(strlen(lxurl)+strlen("/kernel/vxxxx.yyyy/")+strlen(suf)+1);
  slash=lxurl[strlen(lxurl)-1]=='/'?"":"/";
  sprintf(s,"%s%skernel/v%d.%d/%s",lxurl,slash,v[0],v[1],suf);
  dp(DALL,"lkurl: \"%s\"",s);
  return(s);
}

static BUF*dlurl(UBUF*ub,int b,char*suf,int upd) {
 UBUFE*ue=ubel(ub,b);
 BUF*bb;
 int e;
 char*url=emalloc(strlen(ue->url)+strlen(suf)+1);
  sprintf(url,"%s%s",ue->url,suf);
  if((e=dlget(url,ue->vhost,bb=newbuf(),dlto))!=URG_OK) {
    dp(DPRG,"download failed - giving this mirror a penalty");
    cfpen(ue->url,0.3);
    return(0);
  }
  if(upd) cfupd(ue->url,dlq());
  return(bb);
}

static BUF*dltool(UBUF*ub,int b,int v[4],char*suf,int upd) {
 char*p;
  p=alloca(20+strlen(suf));
  sprintf(p,"/kernel/v%d.%d/%s",v[0],v[1],suf);
  return(dlurl(ub,b,p,upd));
}

int apply(UBUF*ub,int b,BUF*ib,int v[4]) {
 static int first=1;
 static char*gzip,*bzip;
 char*s,*prog,cmd[256];
 BUF*bb;
  dp(DPRG,"%d.%d.%d: retrieving-",v[0],v[1],v[2]);
  if(first) {
    first=0;
    gzip=progpath("gzip");
    bzip=progpath("bzip2");
    if(!gzip&&!bzip) error(1,0,"you need gzip or bzip2");
    if(!progpath("patch")) error(1,0,"you need patch");
    dp(DALL,"gzip: %s, bzip2: %s",gzip,bzip);
  }
  if(bzip&&inindex(ib,s=name("patch",v,"bz2"))) { /* get bzip-patch */
    prog="bzip2";
  } else if(gzip&&inindex(ib,s=name("patch",v,"gz"))) { /* get gzip-patch */
    prog="gzip";
  } else { /* no such patch */
    dp(DPRG,"No patch found for %d.%d.%d",v[0],v[1],v[2]);
/*    cfpen(ubel(ub,b)->url,0.2);*/
    ubdelel(ub,b);
    return(0);
  }
  if(!(bb=dltool(ub,b,v,s,1))) return(0);
  dp(DPRG," - applying");
#if 0
  sprintf(cmd,"%s -cd <%s| patch --dry-run -sf -d %s -p1",
    prog,buffile(bb),srcdir);
  if(system(cmd)) {
    error(1,0,"error: patch does not go in cleanly - source left unchanged");
  }
#endif
  sprintf(cmd,"%s -cd <%s| patch -sf -p1 -d %s",
    prog,buffile(bb),srcdir);
  dp(DALL,"executing \"%s\"",cmd);
  if(system(cmd)) error(1,0,"patch did not go in cleanly - stop");
  delbuf(bb);
  return(1);
}

void patch(void) {
 int i,b,try,v[4],iver[4],over[4],lver[4];
 BUF*ib;
 UBUF*ub;
  dp(DTRC,"getpatch()");
  verzero(v); verzero(iver); verzero(lver);
  ub=cfread();
  getiver(srcdir,iver);
  vercp(over,iver);
  dp(DPRG,"installed version: %d.%d.%d",iver[0],iver[1],iver[2]);
  for(try=0;ublen(ub);ubdelel(ub,b),try++) {
    if((b=cfsel(ub))<0) error(1,0,"no sites left to try in config file");
    dp(DPRG,"selected mirror %s",ubel(ub,b)->url);
    if(!(ib=dltool(ub,b,iver,"",0))) continue;
    addbuf(ib,"\0",1);
    if(dlvl()>=DDATA) {
      printf("-------index dump--------------\n");
      bufdump(ib);
    }
    getlv(ib,v);
    dp(DPRG,"Mirror claims latest version is %d.%d.%d",v[0],v[1],v[2]);
    if(verobs(iver,v)||verobs(lver,v)) {
      dp(DPRG,"penalizing mirror for being out of date");
      cfpen(ubel(ub,b)->url,0.4);
      continue;
    }
    if(verlt(lver,v)) vercp(lver,v);
    dp(DTRC,"Latest so far: %d.%d.%d",lver[0],lver[1],lver[2]);
    for(i=0;i<2;i++) {
      if(iver[i]!=lver[i]) {
        error(1,0,"getpatch: installed (%d.%d.%d) and latest (%d.%d.%d) "
                  "differ in lvl %d",iver[0],iver[1],iver[2],
                  lver[0],lver[1],lver[2],i);
      }
    }
    if(iver[2]==lver[2]) {
      if(try+1>=ntries) {
        break;
      }
    } else {
      while(iver[2]++<v[2]) {
        if(!apply(ub,b,ib,iver)) break;
      }
      if(--iver[2]==lver[2]) break;
    }
  }
  if(!verlt(over,lver)&&!verlt(lver,over)) { /* over == lver */
    dp(DPRG,"sorry - nothing new in the %d.%d series",iver[0],iver[1]);
  }
  if(verlt(iver,lver)) {
    dp(DPRG,"sorry - can't update to %d.%d.%d",lver[0],lver[1],lver[2]);
  }
}

char*rematch(char*s,char*re) {
 regex_t preq;
 regmatch_t match[10];
 char*rv=NULL,*p;
  regcomp(&preq,re,REG_EXTENDED|REG_NEWLINE);
  for(p=s;!regexec(&preq,p,NEL(match),match,0);p+=match[0].rm_eo) {
   char*m; int l;
    m=alloca(l=match[0].rm_eo-match[0].rm_so+1);
    strncpy(m,p+match[0].rm_so,l-1); m[l-1]='\0';
    if(!rv||strcmp(rv,m)<0) {
      if(rv) free(rv); rv=strdup(m);
    }
  }
  regfree(&preq);
  return(rv);
}

void query(void) {
 int try,b,iv[4],dv[4],sv[4],cv[4];
 char*ss,*ds,*ssuf,*dsuf;
 UBUF*ub=cfread();
 BUF*bb,*sb,*db,*ib;
  getiver(srcdir,iv);
  printf("Installed version:          %d.%d.%d\n",iv[0],iv[1],iv[2]);
  for(try=0;ublen(ub)&&try<ntries;ubdelel(ub,b),try++) {
    if((b=cfsel(ub))<0) error(1,0,"no sites left to try in config file");
    dp(DPRG,"selected mirror %s",ubel(ub,b)->url);
    /* installed branch */
    if(!(ib=dltool(ub,b,iv,"",0))) continue; addbuf(ib,"\0",1);
    getlv(ib,cv);
    printf("Latest in installed branch: %d.%d.%d\n",cv[0],cv[1],cv[2]);
    if(!(bb=dlurl(ub,b,"/kernel/",0))) continue; addbuf(bb,"\0",1);
    /* stable branch */
    ss=rematch(bufptr(bb),"v[1-9]\\.[02468]");
    ssuf=alloca(strlen("/kernel/")+strlen(ss)+5);
    sprintf(ssuf,"/kernel/%s/",ss);
    if(!(sb=dlurl(ub,b,ssuf,0))) continue; addbuf(sb,"\0",1);
    getlv(sb,sv);
    printf("Latest stable version:      %d.%d.%d\n",sv[0],sv[1],sv[2]);
    /* devel branch */
    ds=rematch(bufptr(bb),"v[1-9]\\.[13579]");
    dsuf=alloca(strlen("/kernel/")+strlen(ds)+5);
    sprintf(dsuf,"/kernel/%s/",ds);
    if(!(db=dlurl(ub,b,dsuf,0))) continue; addbuf(db,"\0",1);
    getlv(db,dv);
    printf("Latest devel version:       %d.%d.%d\n",dv[0],dv[1],dv[2]);
    delbuf(bb); delbuf(sb); delbuf(db);
  }
}
