/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>

#include "lkpatch.h"

extern char*connerr;

/*----------------------------------------------------------------*/

char*urlhost(char*url) { /* extract host from url */
 char*p1,*p2,*rv;
  if(!(p1=strstr(url,"//"))) return(NULL);
  p1+=2;
  if(!(p2=index(p1,'/'))) return(NULL);
  if(!(rv=malloc(p2-p1+1))) return(NULL);
  strncpy(rv,p1,p2-p1);
  rv[p2-p1]='\0';
  return(rv);
}

int urlport(char*url) {
  if(!strncmp(url,"ftp://",6)) return(21);
  if(!strncmp(url,"http://",7)) return(80);
  return(-1);
}

char*urlpath(char*url) {
 char*p;
  if(!(p=strstr(url,"//"))) return(NULL);
  if(!(p=index(p+2,'/'))) return(NULL);
  return(strdup(p));
}

/*----------------------------------------------------------------------*/

void expand(UBUF*sites) { /* expand sites with more than one ip */
 UBUFE ue,*pe;
 int i,j,naddr,n=ublen(sites);
 struct hostent*h;
 u32**ipp;
  for(i=0;i<n;i++) {
    pe=ubel(sites,i);
    dp(DALL,"expand nr %d: %s (%s)\n",i,pe->url,ubgeturl(sites,i));
    pe->host=urlhost(pe->url);
    pe->vhost=estrdup(pe->host);
    pe->port=urlport(pe->url);
    dp(DPRG,"resolving %-20.20s: -",pe->host);
    for(j=0;j<2;j++) {
      if((h=gethostbyname(pe->host))) break;
      if(h_errno!=TRY_AGAIN) break;
    }
    if(!h) {
      dp(DPRG,"can't resolve (skipping)");
      ubdelel(sites,i--); n--;
      continue;
    }
    for(naddr=0,ipp=(u32**)h->h_addr_list;*ipp;ipp++) {
      naddr++;
/*printf("%s\n",dotquad(**ipp));*/
    }
/*fprintf(stderr,"expand: %d addresses\n",naddr);*/
    pe->ip=*(u32*)(h->h_addr_list)[0];
    dp(DPRG,"%16s",dotquad(pe->ip));
    if(naddr>1) {
      /* adjust 1st name */
      pe->path=urlpath(pe->url);
      free(pe->url); pe->url=0;
      free(pe->host); pe->host=0;
      /* additional entries */
      for(j=1;j<naddr;j++) {
        memset(&ue,0,sizeof(ue));
        ue.ip=*(u32*)(h->h_addr_list)[j];
        ue.port=pe->port;
        ue.vhost=estrdup(pe->vhost);
        ue.path=estrdup(pe->path);
        dp(DPRG,"%48s",dotquad(ue.ip));
        ubaddent(sites,&ue);
      }
    }
  }
}

void resip(UBUF*ub) {
 int i,j,n=ublen(ub);
 struct hostent *h1,*h2;
  for(i=0;i<n;i++) {
   UBUFE*pe=ubel(ub,i);
    dp(DPRG,"reverse mapping %s:   -",dotquad0(pe->ip));
    for(j=0;j<3;j++) {
      if((h1=gethostbyaddr((char*)&pe->ip,sizeof(u32),AF_INET))) break;
      if(h_errno!=TRY_AGAIN) break;
    }
    if(!h1) {
      dp(DPRG,"%s (no entry)",dotquad(pe->ip));
      continue;
    }
    for(j=0;j<3;j++) {
      if((h2=gethostbyname(h1->h_name))) break;
      if(h_errno!=TRY_AGAIN) break;
    }
    if(!h2) {
      dp(DPRG,"can't resolve %s (skipping)",h1->h_name);
      continue;
    }
    if( (*(u32*)((h2->h_addr_list)[0])) != pe->ip || (h2->h_addr_list)[1]) {
      dp(DPRG,"(no unique name <-> ip mapping)");
      continue;
    }
    dp(DPRG," %s",h1->h_name);
    pe->host=strdup(h1->h_name);
  }
}

void buildurl(UBUF*ub) {
 int i; char*host;
  for(i=0;i<ublen(ub);i++) {
   UBUFE*pe=ubel(ub,i);
    if(pe->url) continue;
    else if(pe->host) host=pe->host;
    else host=estrdup(dotquad(pe->ip));
    pe->url=malloc(strlen("http://")+strlen(host)+strlen(pe->path)+1);
    switch(pe->port) {
      case 21: strcpy(pe->url,"ftp://"); break;
      case 80: strcpy(pe->url,"http://"); break;
      default: fprintf(stderr,"buildurl: bad port %d\n",pe->port);
    }
    strcat(pe->url,host);
    strcat(pe->url,pe->path);
  }
  for(i=0;i<ublen(ub);i++) dp(DTRC,"buildurl: %d: %s",i,ubel(ub,i)->url);
}

int bestcmp(const void*px,const void *py) {
 long*pa=(long*)px,*pb=(long*)py;
  if(*pa<*pb) return(-1);
  if(*pa>*pb) return(1);
  return(0);
}

void gettimes(UBUF*ub) {
 int i,tmlim=30,tmlim2=30;
 long us,best[NBEST];
  ubsortbyport(ub);
  printf("\nmeasuring connect times ...\n");
  for(i=0;i<NBEST;i++) best[i]=LONG_MAX;
  for(i=ublen(ub)-1;i>=0;i--) {
   UBUFE*pe=ubel(ub,i);
    printf("%s:%02d",dotquad0(pe->ip),pe->port); fflush(stdout);
    if(pe->port==21&&ubexist(ub,pe->ip,80)) {
      pe->q=LONG_MAX;
      printf("  favoring http\n");
      continue;
    }
    pe->q=us=conntm(pe->ip,pe->port,tmlim);
    if(us<=0) {
      pe->q=LONG_MAX;
      printf(" %14s\n",connerr);
      continue;
    }
    printf(" %9ld usec",us);
    if(us<best[NBEST-1]) {
      printf(" *");
      best[NBEST-1]=us;
      qsort(best,NBEST,sizeof(best[0]),bestcmp);
      tmlim2=best[NBEST-1]/1000000+1;
    }
    if(tmlim2<tmlim) tmlim=tmlim2;
    printf("\n");
  }
}

void presel(UBUF*sites) {
 int i;
  dp(DTRC,"presel()");
  expand(sites);
  ubdeldups(sites);
  gettimes(sites);
  ubsortbyq(sites);
  ubtrunc(sites,NBEST);
  resip(sites);
  dp(DPRG,"\nTOP 10:");
  for(i=0;i<ublen(sites);i++) {
   UBUFE*ue=ubel(sites,i);
    printf("%9ld usec %18s:%d   %s\n",
      ue->q,dotquad0(ue->ip),ue->port,ue->host?ue->host:dotquad(ue->ip));
  }
  dp(DPRG,"");
  buildurl(sites);
}
